# Copyright(c) 2022 Association of Universities for Research in Astronomy, Inc.

###############################################################################
# Gemini F2 example data reduction script                                     #
# Typical reduction for: MOS Science and Calibration Data                     #
#                                                                             #
# This script is provided to guide the user through the F2 data reduction     #
# process and may not be optimised to give the best results. It shows the     #
# reduction steps using F2 data and provides explanatory comments at each     #
# step.                                                                       #
#                                                                             #
# It is strongly recommended that the user read the embedded comments and     #
# understand the processing steps, since the optimum steps for a given        #
# dataset may differ, e.g., improved cleaning of cosmic rays and bad pixels   #
# and improved signal-to-noise will most likely be possible. The user can     #
# then edit this script to match their dataset.                               #
#                                                                             #
# Once this script has been properly edited, it can be run by copying and     #
# pasting each command into an IRAF or PyRAF session, or by defining this     #
# script as a task by typing:                                                 #
#                                                                             #
#   ecl> task $thisscript="thisscript.cl"                                     #
#   ecl> thisscript                                                           #
#                                                                             #
# in the IRAF or PyRAF session. It is NOT recommended to run this script      #
# using redirection i.e., cl < thisscript.cl.                                 #
#                                                                             #
# Note that this script is designed to be re-run as needed, so each step is   #
# preceded by a command to delete the files created in that step.             #
#                                                                             #
# The F2 Cookbook at http://gemini-iraf-flamingos-2-cookbook.readthedocs.io/  #
# provides an alternative method for reducing your data in pyRAF              #
###############################################################################

# The data files have been separated by filter and exposure time, where
# appropriate. This information can be found in the primary header unit (PHU)
# of each data file. The imhead or fitsutil.fxhead tasks can be used to view
# the header information in the PHU (or any other extension) of the data file. 
# The hselect task can be used to obtain specific keyword values from the
# headers. Read the help files for these tasks for more information.
#
# Science Dataset:
#
# Observation UT date : 2019 Aug 09
# Data filename prefix: S20190809S
# File numbers:
#   Telluric (HD152602)	         : 83-92   (10x80s)
#   Flats for telluric		 : 93-97   (5x9s)
#   Arc for telluric		 : 98      (80s)
#   Arc lamp off (for background): 99      (80s)
#   Darks for telluric (80s)     : 259-267 (9x80s)
#   Darks for telluric flats (9s): 250-258 (9x9s)
#   Object (S5)                  : 114-121 (8x300s)
#   Flats for object             : 122-125 (4x7s)
#   Arc for object               : 127     (80s)
#   Arc lamp off (for background): 128     (80s)
#   Darks for object (300s)      : 268-276 (9x300s)
#
# Observation UT date : 2019 Aug 11
# Data filename prefix: S20190811S
# File numbers:
#   Darks for objects flats (7s) : 339-345 (7x7s)


###############################################################################
# STEP 1: Initialize the required packages                                    #
###############################################################################

# Load the required packages
gemini
f2

# If copying and pasting these commands into an interactive PyRAF session, use
# the following lines to import the required packages
#from pyraf.iraf import gemini
#from pyraf.iraf import f2

# Use the default parameters except where specified on command lines below
unlearn ("gemini")
unlearn ("f2")
unlearn ("gnirs")
unlearn ("gemtools")

###############################################################################
# STEP 2: Define any variables, the database and the logfile                  #
###############################################################################

# Define any variables (not required if copying and pasting into an interactive
# PyRAF session)
string rawdir, image
struct *scanfile

# Define the logfile
f2.logfile = "f2_mos_example.log"

# To start from scratch, delete the existing logfile
delete (f2.logfile, verify-)

# Define the database directory
f2.database = "f2_mos_database/"

# To start from scratch, delete the existing database files
#
# If copying and pasting these commands into an interactive PyRAF session, use
# something similar to the following example instead of using the uncommented
# lines below.
#
#     >>> if (iraf.access(f2.database)):
#     ...    print "EXAMPLE: Deleting contents of %s" % (f2.database)
#     ...    iraf.delete (f2.database + "*", verify=no)

if (access(f2.database)) {
    printf ("EXAMPLE: Deleting contents of %s\n", f2.database)
    delete (f2.database//"*", verify=no)
}
;

# Define the directory where the raw data is located
# Don't forget the trailing slash!
rawdir = "raw/"
printf ("EXAMPLE: Raw data is located in %s\n", rawdir)

# Load the header keywords for F2
nsheaders ("f2", logfile=f2.logfile)

# Set the display
set stdimage=imt2048

###############################################################################
# STEP 3: Create the reduction lists                                          #
###############################################################################

delete ("*.lis", verify-)

# The user should edit the parameter values in the gemlist calls below to match
# their own dataset
gemlist ("S20190809S", "268-276", > "obj_dark.lis")
gemlist ("S20190809S", "122-125", > "obj_flat.lis")
gemlist ("S20190811S", "339-345", > "obj_flat_dark.lis")
gemlist ("S20190809S", "83-92", > "tel.lis")
gemlist ("S20190809S", "93-97", > "tel_flat.lis")
gemlist ("S20190809S", "250-258", > "tel_flat_dark.lis")
gemlist ("S20190809S", "259-267", > "tel_dark.lis")
gemlist ("S20190809S", "114-121", > "obj.lis")

# We need to isolate the on-source exposures (see Step 16)
print ("S20190809S0114.fits", > "target.lis")
print ("S20190809S0117.fits", >> "target.lis")
print ("S20190809S0118.fits", >> "target.lis")
print ("S20190809S0121.fits", >> "target.lis")

concat ("obj_flat.lis,obj_flat_dark.lis,obj_dark.lis,tel_flat.lis,tel_dark.lis,tel_flat_dark.lis,tel.lis", "calib.lis")
# And the telluric arc and thermal "dark"
print ("S20190809S0098.fits", >> "calib.lis")
print ("S20190809S0099.fits", >> "calib.lis")
# And the MOS arc and thermal "dark"
print ("S20190809S0127.fits", >> "calib.lis")
print ("S20190809S0128.fits", >> "calib.lis")

concat ("calib.lis,obj.lis", "all.lis")

###############################################################################
# STEP 4: Visually inspect the data                                           #
###############################################################################

# Visually inspect all the data. In addition, all data should be visually
# inspected after every processing step. Once the data has been prepared, it is
# recommended to use the syntax [EXTNAME,EXTVER] e.g., [SCI,1], when defining
# the extension.

# Please make sure a display tool (e.g., ds9, ximtool) is already open.

# If copying and pasting these commands into an interactive PyRAF session, use
# something similar to the following example instead of using the uncommented
# lines below.
#
#     >>> file = open("all.lis", "r")
#     >>> for line in file:
#     ...    image = line.strip() + "[1]"
#     ...    iraf.display(rawdir + image, 1)
#     ...    iraf.sleep(5)
#     ...
#     >>> file.close()

scanfile = "all.lis"
while (fscan(scanfile, image) != EOF) {
    display (rawdir//image//"[1]", 1)
    sleep 5
}
scanfile = ""

###############################################################################
# STEP 5: f2prepare all the data                                              #
###############################################################################

# Run F2PREPARE on all the data to update the headers, derive variance and data
# quality (DQ) planes, correct for non-linearity (not yet implemented) and flag
# saturated and non-linear pixels in the DQ plane.

imdelete ("f@all.lis", verify-)
f2prepare ("@all.lis", rawpath=rawdir, fl_vardq+, fl_correct+, \
    fl_saturated+, fl_nonlinear+, mdfdir=rawdir)

###############################################################################
# STEP 6: Create the necessary dark images                                    #
###############################################################################

imdelete ("*dark.fits", verify-)
gemcombine ("f//@obj_dark.lis", "obj_dark.fits", combine="average", \
           fl_vardq+, logfile=f2.logfile)
gemcombine ("f//@obj_flat_dark.lis", "objflat_dark.fits", combine="average", \
	   fl_vardq+, logfile=f2.logfile)
gemcombine ("f//@tel_dark.lis", "tel_dark.fits", combine="average", \
           fl_vardq+, logfile=f2.logfile)
gemcombine ("f//@tel_flat_dark.lis", "telflat_dark.fits", combine="average", \
	   fl_vardq+, logfile=f2.logfile)

###############################################################################
# STEP 7: Create the normalised flat field and BPM for the telluric data      #
###############################################################################

imdelete ("df@tel_flat.lis", verify-)
nsreduce ("f//@tel_flat.lis", outprefix="d", fl_cut+, fl_process_cut-, \
	 fl_dark+, darkimage="telflat_dark.fits", fl_sky-, fl_flat-)

imdelete ("telflat.fits,tel_bpm.pl", verify-)
nsflat ("df//@tel_flat.lis", flatfile="telflat.fits", order=-1, \
       thr_flo=0.1, thr_fup=3.0, bpmfile="tel_bpm.pl")

###############################################################################
# STEP 8: Reduce the arc for the telluric data and determine the wavelength   #
#         solution                                                            #
###############################################################################

# In this example, a single exposure with the arc lamp off has been taken
# for subtraction from the arc, in order to remove the thermal emission.

imdelete ("*tel_arc.fits", verify-)
gemarith ("fS20190809S0098.fits", "-", "fS20190809S0099.fits", "tel_arc.fits", \
	 fl_vardq+, logfile=f2.logfile)
nsreduce ("tel_arc.fits", fl_cut+, fl_flat-, fl_dark-, fl_sky-)

# Now determine the wavelength solution. It is recommended to run nswavelength
# interactively. The default settings work well for most filter / grism
# combinations. However, for Y band data, the following additional parameters
# should be set: threshold=50, nfound=3, nsum=1.

nswavelength ("rtel_arc.fits", nsum=10, threshold=2000, step=5, order=3)

###############################################################################
# STEP 9: Reduce the telluric data                                            #
###############################################################################

# Subtract the dark from the telluric images prior to cutting and flat
# dividing.

imdelete ("df@tel.lis", verify-)
nsreduce ("f//@tel.lis", outprefix="d", fl_cut-, fl_process_cut-, fl_dark+, \
	 darkimage="tel_dark.fits", fl_sky-, fl_flat-)

# Spectra of the telluric were taken at 6 positions along the slit.
# Because the spectral resolution changes with position, better results
# could be obtained by performing the telluric correction with the telluric
# spectrum closest in slit position. In such a case you would need to
# create lists of exposures (and suitable offset exposures) for each
# position, and perhaps set the nsreduce.skyrange parameter to ensure that
# the sky frames are found. However, for most science applications,
# a single telluric correction is sufficient, and that is how we proceed here.

imdelete ("rdf@tel.lis", verify-)
nsreduce ("df//@tel.lis", fl_cut+, fl_dark-, fl_sky+, fl_flat+, \
         flatimage="telflat.fits")

###############################################################################
# STEP 10: Combine the telluric data                                          #
###############################################################################

imdelete ("tel.fits", verify-)
nscombine ("rdf//@tel.lis", output="tel.fits", bpm="tel_bpm.pl", \
          fl_shiftint-, fl_cross+)

display ("tel.fits[SCI,1]", 1)

###############################################################################
# STEP 11: Wavelength calibrate the telluric data                             #
###############################################################################

# The nsfitcoords task is used to determine the combined solution (consisting of
# the wavelength solution) to be applied to the data. The nstransform task is
# used to apply this combined solution. nsfitcoords is best run interactively.

imdelete ("ftel.fits", verify-)
nsfitcoords ("tel.fits", lamptrans="wrtel_arc.fits")

imdelete ("tftel.fits", verify-)
nstransform ("ftel.fits")

display ("tftel.fits[SCI,1]", 2)

###############################################################################
# STEP 12: Extract the telluric spectrum                                      #
###############################################################################

imdelete ("xtftel.fits", verify-)
nsextract ("tftel.fits", fl_apall+, fl_findneg-, fl_trace+)

splot ("xtftel.fits[SCI,1]")

###############################################################################
# We now return to the science frames. The steps are broadly similar to those #
# used for the telluric standard.                                             #
###############################################################################
# STEP 13: Trace the MOS slit edges using the flat field                      #
###############################################################################

# Subtract the dark from the flat images prior to cutting.

imdelete ("df@obj_flat.lis", verify-)
nsreduce ("f//@obj_flat.lis", outprefix="d", fl_cut-, fl_process_cut-, \
	 fl_dark+, darkimage="objflat_dark.fits", fl_sky-, fl_flat-)

# We need to stack the flats to get the best trace of the slit edges
imdelete ("stack.fits", verify-)
gemcombine ("df//@obj_flat.lis", "stack.fits", combine="average", fl_vardq+, \
           logfile=f2.logfile)
f2cut ("stack.fits", outimages="cut_flat.fits", gradimage="stack.fits",
      outslitim="slits.fits")

# Check the cut regions look good

f2display("cut_flat.fits", 1)

###############################################################################
# STEP 14: Create the normalised flat field and BPM for the science data      #
###############################################################################

# We first cut the individual flat field frames using the stack as reference.

imdelete ("cdf@obj_flat.lis", verify-)
f2cut ("df//@obj_flat.lis", gradimage="", refimage="cut_flat.fits")

# The response should look similar for all slits (and similar to the telluric
# flat, since they are taken with the same filter/grism combination).

nsflat ("cdf//@obj_flat.lis", flatfile="objflat.fits", order=-1, \
       thr_flo=0.1, thr_fup=3.0, bpmfile="mos_bpm.pl")
f2display("objflat.fits", 1)

# Determine the s-distortion correction. When running nssdist interactively, a
# window will pop up showing two peaks that correspond to slit edges found in
# the gradient image by f2cut. The expected positions of the peaks (slit edges)
# are printed to screen (under "Coordinate list:"). Press m on one of the two
# visible peaks. The first two of the three numbers that appear at the bottom
# of the plot window indicate the pixel position of peak that was just marked.
# The third of the three numbers (the one in parentheses) will match with one
# of the expected positions of the slit edges. Check that these numbers match
# well (they should be within ~1 pixel) and press return. Repeat for the second
# peak. Press q to move to the next extension.

# For data with well separated slits like this one, it is not necessary to run
# nssdist interactively and the following call works well

nssdist ("slits.fits", nsum=2, fwidth=4, cradius=5, minsep=2)

###############################################################################
# STEP 15: Reduce the arc for the science data and determine the wavelength   #
#          solution                                                           #
###############################################################################

# The quality of the fit of the wavelength solution is improved when the arcs 
# are flat fielded. 

# Subtract the dark from the arc images prior to cutting and flat dividing.

imdelete ("arc.fits", verify-)
gemarith ("fS20190809S0127.fits", "-", "fS20190809S0128.fits", "arc.fits", \
	 fl_vardq+, logfile=f2.logfile)

# Cut the arc images. The quality of the wavelength fit may be improved by
# flatfielding, so the call to "nsreduce" could be changed to "fl_flat+"
# and flatimage="cut_flat.fits" being specified

imdelete ("rarc.fits", verify-)
nsreduce ("arc.fits", fl_cut+, refimage="cut_flat.fits", fl_flat-, \
	 fl_dark-, fl_sky-) 

# Apply the s-distortion correction to the arc

imdelete ("frarc.fits", verify-)
nsfitcoords ("rarc.fits", sdisttransf="slits.fits")

imdelete ("tfrarc.fits", verify-)
nstransform ("frarc.fits")

# Now determine the wavelength solution. It is *strongly* recommended to run
# nswavelength interactively, as there will almost certainly be a small number
# of slits which require user input.

imdelete ("wtfrarc.fits", verify-)
nswavelength("tfrarc.fits", nsum=10, threshold=1000, step=5, \
             order=3, sdist="slits.fits")

###############################################################################
# STEP 16: Reduce and combine the science data                                #
###############################################################################

# Subtract the dark from the science images prior to cutting and flat dividing.

imdelete ("df@obj.lis", verify-)
nsreduce ("f//@obj.lis", outprefix="d", fl_cut-, fl_process_cut-, \
    fl_dark+, darkimage="obj_dark.fits", fl_sky-, fl_flat-)

# We reduce *all* the frames, but only send the on-source frames to nscombine.

imdelete ("rdf@obj.lis", verify-)
nsreduce ("df//@obj.lis", fl_cut+, refimage="cut_flat.fits",
         fl_dark-, fl_flat+, flatimage="objflat.fits", fl_sky+)

imdelete ("obj.fits", verify-)
nscombine ("rdf@target.lis", output="obj.fits", fl_shiftint-, \
    fl_cross-, rejtype="minmax", bpm="mos_bpm.pl")

f2display ("obj.fits", 1)

###############################################################################
# STEP 17: Wavelength calibrate the science data                              #
###############################################################################

# The nsfitcoords task is used to determine the combined solution (consisting of
# the wavelength solution and the spatial distortion solution) to be applied to
# the data. The nstransform task is used to apply this combined solution.
# nsfitcoords is best run interactively.

imdelete ("fobj.fits", verify-)
nsfitcoords ("obj.fits", lamptransf="wtfrarc.fits", \
            sdisttransf="slits.fits")

# The extracted telluric spectrum must be specified as a reference so that
# the transformed slit spectra all have the same wavelength scale.

imdelete ("tfobj.fits", verify-)
nstransform ("fobj.fits", reference="xtftel.fits")

f2display ("tfobj.fits", 2)

###############################################################################
# STEP 18: Extract the science spectrum                                       #
###############################################################################

# When extracting MOS data, nsextract is best run interactively.
imdelete ("xtfobj.fits", verify-)
nsextract ("tfobj.fits", nsum=10, fl_apall+, fl_findneg-, \
          fl_trace+, tr_nsum=10, tr_step=10)

splot ("xtfobj.fits[SCI,16]")

###############################################################################
# STEP 19: Apply the telluric correction to the science spectrum              #
###############################################################################

# Note that this telluric has not been corrected to remove intrinsic stellar
# features; this will leave false emission features in the final spectrum.
# 
# Slit number 16 has very high S/N, so we use this to determine the shift
# and scaling for the telluric correction

imdelete ("axtfobj.fits", verify-)
nstelluric ("xtfobj.fits", "xtftel.fits", ifuextn=16, threshold=0.01)

splot ("axtfobj.fits[SCI,16]")

###############################################################################
# STEP 20: Tidy up                                                            #
###############################################################################

delete ("*.lis", ver-)

###############################################################################
# Finished!                                                                   #
###############################################################################
