# Copyright(c) 2024 Association of Universities for Research in Astronomy, Inc.
#
# GNIRS LR-IFU Calibration and Science data reduction script
#
# Observation UT dates:  August 5, 2022
#
# These data were taken during the commissioning of the GNIRS LR-IFU mode.
# When adapting the example script for other datasets, fl_inter parameters
# should usuallly be enabled in the first instance, to confirm the best values
# for fitting parameters (etc.) interactively, after which those values can
# generally be passed non-interactively, as part of the command line.
#
# Required packages:
#   gemini
#   gnirs
#

# Location of the raw data
set rawdir=/export/data/gemini_testdata/gnirs_testdata/ifu/

# Load required tasks/packages
gemini
gemtools
gnirs

#--------------setup for gnirs-----------
# set default parameters
# use gnirs defaults except where specified on command lines below
unlearn gemini
unlearn gemtools
unlearn gnirs

# set display
set stdimage=imt1024

# set logfile and database directory
gnirs.logfile="GN-2022A-ENG-132.log"
gnirs.database="GN-2022A-ENG-132_database"

# set raw path
nsprepare.rawpath="rawdir$"

# load gnirs header keywords
nsheaders gnirs


###############################################
###### Do basic processing of GCAL flats ######
###############################################
#   Note: INDEF values for shiftx and shifty trigger the automatic
#         cross-correlation algorithm for each set of flats.

gemlist range="116-120" root="N20220805S" > stdflat.lis
gemlist range="97-101" root="N20220805S" > sciflat.lis
nsprepare @stdflat.lis shiftx=INDEF shifty=INDEF
nsprepare @sciflat.lis shiftx=INDEF shifty=INDEF
# shift found: 0.1 pixels in both cases
nsreduce n@stdflat.lis fl_cut+ fl_nsappw- fl_sky- fl_dark- fl_flat-
nsreduce n@sciflat.lis fl_cut+ fl_nsappw- fl_sky- fl_dark- fl_flat-

# The data should be reduced with the closest flat to account
# for any differences due to flexure.  The flats for the standard
# star and the science data are therefore created separately.
#
# Create flat field for telluric standard
nsflat rn@stdflat.lis darks="" flatfile="" darkfile="" fl_save_dark+ \
    process="fit" thr_flo=0.15 thr_fup=2.5 fl_corner+

# Create flat field for the science frames
nsflat rn@sciflat.lis darks="" flatfile="" darkfile="" fl_save_dark+ \
    process="fit" thr_flo=0.15 thr_fup=2.5 fl_corner+

################################################################
##### Trace the spatial tilt/curvature in the flat fields  #####
################################################################

# For the telluric standard flat
nfflt2pin nN20220805S0116 outimage=nN20220805S0116_pin
nscut nN20220805S0116_pin
nssdist snN20220805S0116_pin fwidth=20 thresh=20 fl_inter-

# For the science flat
nfflt2pin nN20220805S0097 outimage=nN20220805S0097_pin
nscut nN20220805S0097_pin
nssdist snN20220805S0097_pin fwidth=20 thresh=20 fl_inter-


##############################################################
##### Reduce the arcs and determine wavelength solutions #####
##############################################################
#    The user might edit the default line list and/or additional
#    fitting parameters for optimal results.

gemlist range="121-122" root="N20220805S" > stdarc.lis
nsprepare @stdarc.lis shiftimage="nN20220805S0116" \
    bpm="rnN20220805S0116_flat_bpm.pl"
nsreduce n@stdarc.lis fl_cut+ fl_nsappw+ fl_sky- fl_dark- fl_flat-
nswavelength rn@stdarc.lis nsum=13 step=13 nlost=7 nfound=6 match=-5 \
    thresh=100 low_rej=2.5 high_rej=2.5 fl_inter+

gemlist range="102-103" root="N20220805S" > sciarc.lis
nsprepare @sciarc.lis shiftimage="nN20220805S0097" \
    bpm="rnN20220805S0097_flat_bpm.pl"
nsreduce n@sciarc.lis fl_cut+ fl_nsappw+ fl_sky- fl_dark- fl_flat-
nswavelength rn@sciarc.lis nsum=13 step=13 nlost=7 nfound=6 match=-5 \
    thresh=100 low_rej=2.5 high_rej=2.5 fl_inter+


###########################################################
###### Restore slice-to-slice variations in the flats #####
###########################################################
#   Here the same GCal flat is used as in nsflat, but we could use any
#   spatially-flat reference, such as a twilight spectrum, as long as the
#   flexure (& hence illumination) is similar to that of the science data.
#   The exposures are re-processed with partial flat fielding & wavelength
#   rectification in order to measure relative slice levels accurately.

nsreduce nN20220805S0116 fl_cut+ fl_nsappw+ fl_sky- fl_dark- fl_flat+ \
    flatimage="rnN20220805S0116_flat" outprefix="g"
nsreduce nN20220805S0097 fl_cut+ fl_nsappw+ fl_sky- fl_dark- fl_flat+ \
    flatimage="rnN20220805S0097_flat" outprefix="g"

nsfitcoords gnN20220805S0116 lamp=wrnN20220805S0121 \
    sdist=snN20220805S0116_pin database=GN-2022A-ENG-132_database
nsfitcoords gnN20220805S0097 lamp=wrnN20220805S0102 \
    sdist=snN20220805S0097_pin database=GN-2022A-ENG-132_database

nstransform fgnN20220805S0116 database=GN-2022A-ENG-132_database
nstransform fgnN20220805S0097 database=GN-2022A-ENG-132_database

# Here a suitable statistics section has been determined by inspection.
nfnormslices tfgnN20220805S0116 rnN20220805S0116_gflat \
    flat=rnN20220805S0116_flat statsec="[*,150:950]"

nfnormslices tfgnN20220805S0097 rnN20220805S0097_gflat \
    flat=rnN20220805S0097_flat statsec="[*,100:900]"


##############################################
##### Reduce the standard star exposures #####
##############################################

# These data were taken as an A B A' sequence with a nod to sky; specify
# the object & sky lists separately, to reduce only the exposures we want.
# The shiftimage parameter is set to one of the GCAL flat images that
# were used to create the flat field; the MDF shift to apply will be taken
# from that image.
gemlist range="110-115" root="N20220805S" > std.lis
gemlist range="110,112-113,115" root="N20220805S" > stdobj.lis
nsprepare @std.lis shiftimage="nN20220805S0116" \
    bpm="rnN20220805S0116_flat_bpm.pl"
nsreduce n@stdobj.lis fl_cut+ fl_nsappwave+ fl_dark- fl_sky+ fl_flat+ \
    skyimages="nN20220805S0111,nN20220805S0111,\
               nN20220805S0114,nN20220805S0114" \
    flatimage="rnN20220805S0116_gflat"
nsfitcoords rn@stdobj.lis lamp=wrnN20220805S0121 \
    sdist=snN20220805S0116_pin database=GN-2022A-ENG-132_database
nstransform frn@stdobj.lis database=GN-2022A-ENG-132_database

# Here nsstack could optionally be used to combine exposures taken at the
# same spatial dither position before continuing as below.

# Extract a 1D spectrum for each slice, processing the exposures at spatial
# dither position with an appropriate background sample region. Spurious
# peak identifications / traces in the outer, noisy slices do not matter,
# as those with very low signal get discarded below. Should absolute fluxes
# be required, it would be important to review the fitting parameters
# interactively, to ensure that all of the signal gets captured (this can be
# checked using nfapsum with no prior nsextract, as a cruder extraction).
nsextract tfrnN20220805S0110,tfrnN20220805S0113 \
    database=GN-2022A-ENG-132_database ylevel=0.01 bgsample="-16:-8,8:16" \
    tr_order=4 tr_niter=3 tr_lowrej=2.5 tr_highrej=2.5 fl_inter-
nsextract tfrnN20220805S0112,tfrnN20220805S0115 \
    database=GN-2022A-ENG-132_database ylevel=0.01 bgsample="-18:-9,8:14" \
    tr_order=4 tr_niter=3 fl_inter-

# The range of adequately-illuminated slices to use here for each dither
# position has been pre-determined by manual inspection (the default of all
# slices could be used, but would be noisy).
nfapsum xtfrnN20220805S0110,xtfrnN20220805S0113 extrange="9-17"
nfapsum xtfrnN20220805S0112,xtfrnN20220805S0115 extrange="7-15"
nfapsum ixtfrnN20220805S0110,ixtfrnN20220805S0112 \
    outspectrum="ixtfrnN20220805S0110_tell"


####################################################
##### Basic reduction of the science exposures #####
####################################################

# The shiftimage parameter is set to one of the GCAL flat images that
# were used to create the flat field; the MDF shift to apply will be taken
# from that image.
gemlist range="85-96" root="N20220805S" > sci.lis
nsprepare @sci.lis shiftimage="nN20220805S0097" \
    bpm="rnN20220805S0097_flat_bpm.pl"

# The science observations have been taken in an AAA BBB B'B'B' A'A'A'
# sequence with a nod to sky, so combine the triplets at each position with
# rejection before subtracting the sky etc. Sigma clipping must be done
# conservatively, to avoid masking the bright source, due to slight changes
# in seeing, pointing, transparency etc. Some optional absolute rejection
# thresholds have been pre-determined by visual inspection of the brightness
# range of real features in the data. Unfortunately, nsstack regenerates the
# VAR extensions based on pixel statistics, producing values an order of
# magnitude greater, which could be avoided by reducing pairs of exposures
# and stacking the resulting 6 datacubes with an external program instead.
nsstack nN20220805S0085,nN20220805S0086,nN20220805S0087 tolerance=0.1 \
    combtype="average" rejtype="avsigclip" lthreshold=-50 hthreshold=500. \
    nkeep=1 lsigma=7. hsigma=7.
nsstack nN20220805S0088,nN20220805S0089,nN20220805S0090 tolerance=0.1 \
    combtype="average" rejtype="avsigclip" lthreshold=-50 hthreshold=150. \
    nkeep=1 lsigma=7. hsigma=7.
nsstack nN20220805S0091,nN20220805S0092,nN20220805S0093 tolerance=0.1 \
    combtype="average" rejtype="avsigclip" lthreshold=-50 hthreshold=150. \
    nkeep=1 lsigma=7. hsigma=7.
nsstack nN20220805S0094,nN20220805S0095,nN20220805S0096 tolerance=0.1 \
    combtype="average" rejtype="avsigclip" lthreshold=-50. hthreshold=550. \
    nkeep=1 lsigma=7. hsigma=7.

# Complete the basic reduction. Some sky subtraction residuals remain in
# exposure 94, which might need masking by the user.
nsreduce "nN20220805S0085_stack,nN20220805S0094_stack" \
    fl_cut+ fl_nsappwave+ fl_dark- fl_sky+ fl_flat+ \
    skyimages="nN20220805S0088_stack,nN20220805S0091_stack" \
    flatimage="rnN20220805S0097_gflat"


# ###################################################
# ##### Resample the spectra onto a linear grid #####
# ###################################################

nsfitcoords rnN20220805S0085_stack,rnN20220805S0094_stack, \
    lamp=wrnN20220805S0102 sdist=snN20220805S0097_pin \
    database=GN-2022A-ENG-132_database

nstransform frnN20220805S0085_stack,frnN20220805S0094_stack \
    database=GN-2022A-ENG-132_database


#########################################################
##### Apply the telluric correction to science data #####
#########################################################

# Interpolate over the Brackett Gamma feature in the standard (using the
# same fitting parameters as nstelluric below), so it won't produce a spike
# adjoining the real H2 feature in this particular science dataset.
fit1d ixtfrnN20220805S0110_tell[sci] ixtfrnN20220805S0110_fit fit bpm="" \
    sample="*" nav=1 func=spline3 order=10 low_rej=1.5 high_rej=3 niter=3 \
    grow=1 inter-
# The image section to replace with fit values here was determined by manual
# inspection of the telluric spectrum.
imcopy ixtfrnN20220805S0110_fit[423:443] \
    ixtfrnN20220805S0110_tell[sci][423:443]

# Determine a WCS shift between science & telluric data after resampling,
# to align the spectra more reliably than the xcorr option in nstelluric.
wcsmap tfrnN20220805S0085_stack[sci,11][24,*] ixtfrnN20220805S0110_tell[sci] \
    aligndb fitgeom=shift verb+ inter- |match shift

nstelluric tfrnN20220805S0085_stack,tfrnN20220805S0094_stack \
    ixtfrnN20220805S0110_tell fl_xcorr- fl_tweak- threshold=0.01 \
    shift=30.88 niter=3 fl_inter-


##################################################
##### Stack the 2D spectra into 3D datacubes #####
##################################################

nfcube atfrnN20220805S0085_stack,atfrnN20220805S0094_stack process="rotate"

