# Copyright(c) 2004-2022 Association of Universities for Research in Astronomy, Inc.

procedure nxdisplay (image, frame)
 
# Quick and dirty procedure to display all extensions of a processed GNIRS XD or IFU image

# 2004-Jan-30, cwinge, original version
# 2017-Mar-01, astephens, tighten gaps between orders and remove y-magnification
# 2017-Dec-29, astephens, catch incorrect NSCIEXT values
# 2021-Jul-25, astephens, allow specifying extension name (SCI, DQ, VAR)

string    image    {prompt="Raw image"}
int       frame    {1, prompt = "Frame to be written to"}
char      extname  {"SCI", enum="SCI|DQ|VAR", prompt="Extension name"}
real      z1       {INDEF, prompt="Minimum level to be displayed"}
real      z2       {INDEF, prompt="Maximum level to be displayed"}
bool      verbose  {no,            prompt="Verbose debugging output"}
char      sversion {"2021-Jul-25", prompt="Software version"}

begin

    string l_image, grating, extn
    bool   erase_frame, zscale, zrange
    int    junk, l_frame, num_sci_ext, ii, status
    int    x_width, total_width, sci_width, disp_width, x_border
    real   x_center, x_offset, x_start, x_scale

    junk = fscan (image, l_image)
    l_frame = frame
    status = 0                                                 # Define default status    
    keypar (l_image//"[0]", "NSCIEXT", silent+)                # Read number of sci extensions
    if (!keypar.found) {
        print ("NXDISPLAY ERROR - "//l_image//" has not been processed."//\
            " Cannot display image. Exiting.")
        goto crash
    } else {
        num_sci_ext = int(keypar.value)
        if (verbose) print("...num_sci_ext = ", num_sci_ext)
    }

    # If z1 and z2 are supplied, turn off zscale and zrange:
    if ( (z1 != INDEF) && (z2 != INDEF) ) {
      zscale = no
      zrange = no
    } else if ( (z1 != INDEF) || (z2 != INDEF) ) {
      print ("ERROR:  Please specify both z1 and z2 for manual zrange")
      goto clean
    } else {
      zscale = yes
      zrange = yes
    }

    # Verify the image has all the SCI extensions it claims it does
    # (nsflat crashes result in an incorrect NSCI header value)
    fxhead (l_image) | fields ("STDIN", 3) | match ("SCI", "STDIN", stop-) | count | scan (junk)
    if (verbose) print ("...found ", junk, "SCI extensions")
    if (num_sci_ext != junk) {
        print ("WARNING: ", l_image, " is missing extensions")
        print ("(the header claims ", num_sci_ext, "but ", junk, "were found)")
        num_sci_ext = junk
    }

    sci_width = 0
    for (ii = 1; ii <= num_sci_ext; ii += 1) {                 # Loop over SCI extensions
        extn = "["//extname//","//ii//"]"
        keypar (l_image//extn, "i_naxis1", silent+)            # Read the NAXIS2 keyword
        x_width = int(keypar.value)
        sci_width += x_width                                   # Accumulate the science width
    }

    if (verbose) print("...science width = ", sci_width)
    x_border = sci_width / 200                                 # Border size (pix)
    if (verbose) print("...border width = ", x_border, "pix")
    total_width = sci_width + (num_sci_ext - 1) * x_border     # Total width (pix)
    if (verbose) print("...total width = ", total_width, "pix")
    disp_width = 2048                                          # Display width (pix)
    set stdimage = "imt" // disp_width                         # Set the display size
    x_start = (disp_width - total_width)/2. / real(disp_width) # Fraction of image display

    for (ii = 1; ii <= num_sci_ext; ii += 1) {                 # Loop over SCI extensions
        extn = "["//extname//","//ii//"]"                      # Set the extension to display
        keypar (l_image//extn, "i_naxis1", silent+)            # Read the width of the extension
        if (verbose) print("...x_start = ", x_start)
        x_width = int(keypar.value)
        if (verbose) print("...x_width = ", x_width, "pix")
        x_center = x_start + (x_width/2 / real(disp_width))
        if (verbose) print("...x_center = ", x_center)
        if (ii == 1) { erase_frame = yes } else { erase_frame = no }
        display (l_image//extn, l_frame, bpdisplay="none", overlay="", \
            erase=erase_frame, border_erase=no, select_frame=yes, \
            repeat=no, fill=no,  zscale=zscale, contrast=0.25, zrange=zrange, \
            nsample=500, xcenter=x_center, \
            ycenter=0.5, xsize=1., ysize=1., xmag=1.0, ymag=1.0, order=0, \
            z1=z1, z2=z2, ztrans="linear", lutfile="")
        x_start = x_start + (x_width + x_border) / real(disp_width)
    }

    goto clean

crash:
    status = 1

clean:

    if (status != 1) {
        status = 0
    }

end
